/*
 * Copyright (c) 2007 The Akuma Project
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * $Id: rstack.c 133 2007-05-29 14:32:04Z asmodai $
 */
/*
    		 Crown Copyright (c) 1997

    This TenDRA(r) Computer Program is subject to Copyright
    owned by the United Kingdom Secretary of State for Defence
    acting through the Defence Evaluation and Research Agency
    (DERA).  It is made available to Recipients with a
    royalty-free licence for its use, reproduction, transfer
    to other parties and amendment for any purpose not excluding
    product development provided that any such use et cetera
    shall be deemed to be acceptance of the following conditions:-

        (1) Its Recipients shall ensure that this Notice is
        reproduced upon any copies or amended versions of it;

        (2) Any amended version of it shall be clearly marked to
        show both the nature of and the organisation responsible
        for the relevant amendment or amendments;

        (3) Its onward transfer from a recipient to another
        party shall be deemed to be that party's acceptance of
        these conditions;

        (4) DERA gives no warranty or assurance as to its
        quality or suitability for any purpose and DERA accepts
        no liability whatsoever in relation to any use to which
        it may be put.
*/


/*** rstack.c --- Renaming stack ADT.
 *
 ** Author: Steve Folkes <smf@hermes.mod.uk>
 *
 *** Commentary:
 *
 * This file implements the renaming stack routines.  They are mainly used by
 * the output routines to do scoping of inlined rules.
 *
 *** Change Log:
 * $Log: rstack.c,v $
 * Revision 1.1.1.1  1998/01/17  15:57:46  release
 * First version to be checked into rolling release.
 *
 * Revision 1.2  1994/12/15  09:58:31  smf
 * Brought into line with OSSG C Coding Standards Document, as per
 * "CR94_178.sid+tld-update".
 *
 * Revision 1.1.1.1  1994/07/25  16:04:37  smf
 * Initial import of SID 1.8 non shared files.
 *
**/

/****************************************************************************/

#include "rstack.h"
#include "action.h"
#include "basic.h"
#include "name.h"
#include "rule.h"
#include "type.h"

/*--------------------------------------------------------------------------*/

void
rstack_init(RStackP rstack)
{
    rstack->head = NIL(TransStackEntryP);
}

void
rstack_push_frame(RStackP rstack)
{
    TransStackEntryP frame = ALLOCATE(TransStackEntryT);

    frame->next = rstack->head;
    rtrans_init(&(frame->translator));
    rstack->head = frame;
}

void
rstack_compute_formal_renaming(RStackP rstack, TypeTupleP names)
{
    ASSERT(rstack->head);
    types_compute_formal_renaming(names, &(rstack->head->translator));
}

void
rstack_compute_formal_inlining(RStackP rstack, TypeTupleP names,
			       TypeTupleP renames)
{
    SaveRStackT state;

    ASSERT(rstack->head);
    state.head = rstack->head->next;
    types_compute_formal_inlining(names, renames, &(rstack->head->translator),
				  &state);
}

void
rstack_compute_local_renaming(RStackP rstack, TypeTupleP names,
			      TypeTupleP exclude, TableP table)
{
    SaveRStackT state;

    ASSERT(rstack->head);
    state.head = rstack->head->next;
    types_compute_local_renaming(names, exclude, &(rstack->head->translator),
				 &state, table);
}

void
rstack_add_translation(RStackP rstack, EntryP from, EntryP to, EntryP type,
		       BoolT reference)
{
    ASSERT(rstack->head);
    rtrans_add_translation(&(rstack->head->translator), from, to, type,
			   reference);
}

void
rstack_save_state(RStackP rstack, SaveRStackP state)
{
    state->head = rstack->head;
}

EntryP
rstack_get_translation(SaveRStackP state, EntryP entry, EntryP *type_ref,
		       BoolT *reference_ref)
{
    TransStackEntryP frame = state->head;

    while (frame) {
	EntryP translation;

	translation = rtrans_get_translation(&(frame->translator), entry,
					     type_ref, reference_ref);
	if (translation) {
	    return(translation);
	}
	frame = frame->next;
    }
    return(NIL(EntryP));
}

void
rstack_apply_for_non_locals(RStackP non_local_stack, SaveRStackP state,
			    void (*proc)(EntryP, EntryP, GenericP),
			    GenericP closure)
{
    TransStackEntryP frame = non_local_stack->head;

    if ((frame != NIL(TransStackEntryP)) && (state->head)) {
	TransStackEntryP limit = state->head->next;

	for (; frame != limit; frame = frame->next) {
	    rtrans_apply_for_non_locals(&(frame->translator), proc, closure);
	}
    }
}

void
rstack_pop_frame(RStackP rstack)
{
    TransStackEntryP frame = rstack->head;

    rstack->head = frame->next;
    rtrans_destroy(&(frame->translator));
    DEALLOCATE(frame);
}

void
rstack_destroy(RStackP rstack)
{
    while (rstack->head) {
	rstack_pop_frame(rstack);
    }
}

/*
 * Local variables(smf):
 * eval: (include::add-path-entry "../os-interface" "../library")
 * eval: (include::add-path-entry "../generated")
 * end:
**/
